#include <stdlib.h>
#include <stdio.h>
#include <libgen.h>
#include <string.h>
#include <linux/version.h>
#include <ctype.h>


#define MAXKEYS 512
#define MAXKEYVAL 200
#define HASHSIZE 101
enum {
	is_shift = 1, is_input, is_redef, is_spk, is_rd
};
typedef struct st_key t_key;
struct st_key {
	char *name;
	t_key *next;
	int value, state;
};
unsigned char key_blank[16], key_data[MAXKEYVAL][16], *kp;

t_key key_table[MAXKEYS];
t_key *extra_keys = key_table+HASHSIZE;
char buffer[256], filename[256];
FILE *infile;
char delims[] = "\t\n ,";
char *dir_name, *def_name, *def_val, *cp;
int map_ver = 121; /* an arbitrary number so speakup can check */
int lc, shift_table[17];
int value, shift_state, i, spk_val = 0, lock_val = 0;
int max_states = 0, max_key_used = 0, num_keys_used = 0;
struct st_key_init {
      char *name;
      int value, state;
};
struct st_key_init  init_keys[] = {
	{ "shift", 1, is_shift  }, { "altgr", 2, is_shift  },
	{ "ctrl",	4, is_shift  }, { "alt",	 8, is_shift  },
	{ "spk", 16, is_shift  }, { "double", 32, is_shift  }
};
struct st_key_init  readall_check[] = {
	{ "key_up", 0, is_rd }, { "key_down", 0, is_rd },
	{ "key_left", 0, is_rd }, { "key_right", 0, is_rd },
	{ "key_pageup", 0, is_rd }, { "key_pagedown", 0, is_rd },
	{ "key_backspace", 0, is_rd }
};

void open_input( char *name )
{
	if ( strchr( name, '/' ) )
		sprintf( filename, "%s/%s", dir_name, name );
	else strcpy( filename, name );
	if ( ( infile = fopen( filename, "r" ) ) == 0 ) {
		fprintf( stderr, "can't open %s\n", filename);
		exit( 1 );
	}
	lc = 0;
}

int
oops( char *msg, char *info )
{
	if ( info == NULL ) info = " ";
	fprintf( stderr, "error: file %s line %d\n", filename, lc );
	fprintf( stderr, "%s %s\n", msg, info );
	exit( 1 );
}

t_key *hash_name( char *name )
{
	u_char *pn = (u_char *)name;
	int hash = 0;
	while ( *pn ) {
			hash = ( hash * 17 ) & 0xfffffff;
		if ( isupper( *pn ) ) *pn = tolower( *pn );
		hash += ( int )*pn;
		pn++;
	}
	hash %= HASHSIZE;
	return &key_table[hash];
}

t_key *find_key( char *name )
{
	t_key *this = hash_name( name );
	while ( this ) {
		if ( this->name && !strcmp( name, this->name ) ) return this;
		this = this->next;
	}
	return this;
}

t_key *add_key( char *name, int value, int state )
{
	t_key *this = hash_name( name );
	if ( extra_keys-key_table >= MAXKEYS )
		oops( "out of key table space, enlarge MAXKEYS", NULL );
	if ( this->name) {
		while ( this->next ) {
			if ( !strcmp( name, this->name ) )
				oops( "attempt to add duplicate key", name );
			this = this->next;
		}
		this->next = extra_keys++;
		this = this->next;
	}
	this->name = strdup( name );
	this->value = value;
	this->state = state;
	return this;
}

int get_defines( )
{
	t_key *prevdef;
	while ( fgets( buffer, 250, infile ) ) {
		lc++;
		if ( strncmp( buffer, "#define", 7 ) ) continue;
		strtok( buffer, delims );
		def_name = strtok( 0, delims );
		if ( strncmp( def_name, "KEY_", 4 ) ) continue;
		def_val = strtok( 0, delims );
		if ( !def_val) oops( "null value", def_name);
		if ( isdigit( *def_val ) ) {
			value = strtol( def_val, NULL, 0 );
			if ( value > 0 && value < MAXKEYVAL )
				add_key(	def_name, value, is_input );
		} else if ( (prevdef = find_key( def_val ))) {
			value = prevdef->value;
			add_key(	def_name, value, is_input );
		} else
		fprintf(stderr, "input key definition rejected %s %s\n",
			def_name, def_val );
	}
	fclose( infile );
	infile = 0;
	return 0;
}

int get_enums( )
{
	int in_enum = 0, value = 0;
	char *cp;
	while ( fgets( buffer, 250, infile ) ) {
				lc++;
				if ( !strncmp( buffer, "enum {", 5 ) ) {
			in_enum = 1;
			continue;
		}
		if ('}' == *buffer ) in_enum = 0;
		if ( !in_enum ) continue;
		def_name = strtok( buffer, delims );
		if ( !def_name || '/' == *def_name ) continue;
		cp = def_name+strlen(def_name)+1;
		if ( *cp == '=' ) {
			while ( *cp && !isdigit(*cp)) cp++;
			value = atoi( cp );
		}
		while (def_name && *def_name != '/') {
			if ( def_name != cp && *def_name >= 'A' )
			add_key(	def_name, value++, is_spk );
			def_name = strtok( 0, delims );
		}
	}
	fclose( infile );
	infile = 0;
	return 0;
}

int get_shift_value( int state )
{
	int i;
	for ( i = 0; i <= max_states; i++ )
		if (shift_table[i] == state) return i;
	if ( i >= 15 )
		oops( "too many shift states", NULL );
	shift_table[i] = state;
	max_states = i;
	return i;
}

int
main( int argc, char *argv[] )
{
	t_key *this;
char *argname = "speakupmap.map", *out_name = "speakupmap.h";
	int bInclude = 0;
	FILE *f_out = stdout;

	bzero( key_table, sizeof( key_table ) );
	bzero( key_blank, sizeof( key_blank ) );
	bzero( key_data, sizeof( key_data ) );
	bzero( shift_table, sizeof( shift_table ) );
	dir_name = getenv( "TOPDIR" );
	if ( !dir_name ) dir_name = "/usr/src/linux";
	for (i = 1; i < argc; i++) {
		if (!strcmp(argv[i], "-i"))
			bInclude = 1;
		else argname = argv[i];
	}
	for ( i = 0; i < 6; i++ )
		add_key( init_keys[i].name, init_keys[i].value, init_keys[i].state );
	open_input( "include/uapi/linux/input-event-codes.h");
	get_defines( );
	open_input( "spk_priv_keyinfo.h" );
	get_enums( );
	this = find_key( "spk_key" );
	if ( this ) spk_val = this->value;
	this = find_key( "spk_lock" );
	if ( this ) lock_val = this->value;
	open_input( argname );
	if (bInclude) {
		if ( ( f_out = fopen( out_name, "w" ) ) == 0 ) {
			fprintf( stderr, "can't open %s\n", out_name);
			exit( 1 );
		}
	}
	while ( fgets( buffer, 250, infile ) ) {
		lc++;
		value = shift_state = 0;
		cp = strtok( buffer, delims );
		if ( *cp == '#' || *cp == '/' ) continue;
		while ( cp && *cp != '=') {
			this = find_key( cp );
			if ( this == NULL )
				oops( "unknown key/modifier", cp );
			if ( this->state == is_shift ) {
				if ( value )
					oops( "modifiers must come first", cp );
				shift_state += this->value;
			} else if ( this->state == is_input || this->state == is_redef )
				value = this->value;
			else oops( "bad modifier or key", cp );
			cp = strtok( 0, delims );
		}
		if ( !cp ) oops( "no = found", NULL );
		cp = strtok( 0, delims );
		if ( !cp ) oops( "no speakup function after =", NULL );
		this = find_key( cp );
		if ( this == NULL || this->state != is_spk )
			oops( "invalid speakup function", cp );
		i = get_shift_value( shift_state );
		kp = key_data[value];
		if ( kp[i] ) {
			while ( --cp > buffer )
				if ( !*cp ) *cp = ' ';
			oops( "two functions on same key combination", cp );
		}
		kp[i] = (char)this->value;
		if ( value > max_key_used ) max_key_used = value;
		if ( ( shift_state&16 ) || kp[i] != spk_val&& kp[i] != lock_val )
			continue;
		// if speakup or spk_lock, and not with speakup key,
		// add with speakup key states so we can get back out of speakup key mode
		shift_state = get_shift_value( shift_state+16 );
		kp[shift_state] = kp[i];
	}
	fclose( infile );
	max_states++; // add state for speakup key down with modifiers
	this = find_key( "cursor_check" );
	if ( this ) {
		int rd_value = this->value;
		// for cursor keys, fill in the unused states for ctl-shift etc.
		for (lc = 0; lc < 7; lc++) {
			if ( !(this = find_key( readall_check[lc].name))) continue;
			value = this->value;
			kp = key_data[value];
			if ( value > max_key_used ) max_key_used = value;
			for ( i = 0; i < max_states; i++ ) {
				if (shift_table[i] >= 32) continue; // no double press states!
				if ( !kp[i]) kp[i] = rd_value;
			}
		}
	}
	for ( lc = 0; lc <= max_key_used; lc++ ) {
		kp = key_data[lc];
		if ( !memcmp( key_blank, kp, 16 ) ) continue;
		num_keys_used++;
		if ( kp[0] != spk_val ) continue;
		// for all modifiers with speakup states,  add speakup_key
		for ( i = 1; i <= max_states; i++ )
			if ( 0 == kp[i]  )
				kp[i] = spk_val;
	}
	max_states++;
	if (bInclude) fprintf(f_out, "{\n");
	fprintf(f_out, "\t%d, %d, %d,\n\t", map_ver, num_keys_used, max_states );
	for ( i = 0; i < max_states; i++ )
				fprintf(f_out, "%d, ", shift_table[i] );
	fprintf(f_out, "0,");
	for ( lc = 0; lc <= max_key_used; lc++ ) {
		kp = key_data[lc];
		if ( !memcmp( key_blank, kp, 16 ) ) continue;
		fprintf(f_out, "\n\t%d,", lc );
		for ( i = 0; i < max_states; i++ )
			fprintf(f_out, " %d,", (unsigned int)kp[i] );
	}
	if (bInclude) {
		fprintf(f_out, "\n\t0, %d\n};\n", map_ver );
		i = (num_keys_used+2)*(max_states+1);
		i = (i+16)&~15;
		fprintf(f_out,"#define KEY_BUFF_SIZE %d\n", i);
	} else fprintf(f_out, "\n\t0, %d\n", map_ver );
	fclose(f_out);
	exit( 0 );
}
